/**
  *  \file planetscentral/talk/cmdthread.cc
  *  \brief Talk: THREAD commands
  */

#include "planetscentral/talk/cmdthread.h"
#include "int/args.h"
#include "int/array.h"
#include "planetscentral/dbclient/connection.h"
#include "planetscentral/dbclient/reconnect.h"
#include "planetscentral/talk/cmdforum.h"
#include "planetscentral/talk/cmdpost.h"
#include "planetscentral/talk/talk.h"
#include "planetscentral/talk/talkconnection.h"
#include "planetscentral/talk/topic.h"
#include "planetscentral/talk/forum.h"
#include "planetscentral/talk/db.h"
#include "planetscentral/talk/message.h"

void
doThreadStat(TalkConnection& conn, IntArgBlock& args)
{
    args.checkArgumentCount(1);
    int32_t tid = toIntResult(args.getNext());

    // Do it
    ReconnectingGuard g(database_rc);
    Topic t(database_connection, tid);
    if (!t.exists()) {
        conn.sendError("404 Not found");
        return;
    }
    PostAccessChecker(conn).checkTopic(t);
    std::auto_ptr<IntValue> result(t.describe());
    conn.sendObject(result.get());
}

void
doThreadMStat(TalkConnection& conn, IntArgBlock& args)
{
    // Do it
    ReconnectingGuard g(database_rc);
    IntDataSegment dseg;
    PostAccessChecker checker(conn);
    while (args.getNumArgs() > 0) {
        int32_t topicId = toIntResult(args.getNext());
        Topic t(database_connection, topicId);
        if (!t.exists() || !checker.isAllowed(t)) {
            dseg.pushNew(0);
        } else {
            dseg.pushNew(t.describe());
        }
    }
    conn.sendArray(dseg, dseg.size());
}

void
doThreadListPosts(TalkConnection& conn, IntArgBlock& args)
{
    // Parse
    ListParams parms;
    parms.parse(args, true, ListParams::postSortKeys);

    // Do it
    ReconnectingGuard g(database_rc);
    Topic t(database_connection, parms.getId());
    if (!t.exists()) {
        conn.sendError("404 Not found");
    } else {
        parms.exec(conn, t.messages());
    }
}

void
doThreadSticky(TalkConnection& conn, IntArgBlock& args)
{
    // Parse
    args.checkArgumentCount(2);
    int32_t topicId = toIntResult(args.getNext());
    int32_t value   = toIntResult(args.getNext());

    // Do it
    ReconnectingGuard g(database_rc);
    Topic t(database_connection, topicId);
    if (!t.exists()) {
        conn.sendError("404 Not found");
        return;
    }

    // Permission check
    conn.checkPrivilege(t.forum().deletePermissions().get());

    // Execute
    t.setSticky(value != 0);
    conn.sendSuccess("OK");
}

void
doThreadPermissions(TalkConnection& conn, IntArgBlock& args)
{
    // Parse args
    args.checkArgumentCountAtLeast(1);
    int32_t topicId = toIntResult(args.getNext());
    int32_t result = 0;
    int32_t mask = 1;

    // Do it
    ReconnectingGuard g(database_rc);
    Topic t(database_connection, topicId);
    if (!t.exists()) {
        conn.sendError("404 Not found");
    } else {
        Forum f = t.forum();
        while (args.getNumArgs() > 0) {
            string_t key = toStringResult(args.getNext()) + "perm";
            string_t str = t.header().getStringField(key);
            if (str.empty()) {
                str = f.header().getStringField(key);
            }
            if (conn.hasPrivilege(str)) {
                result |= mask;
            }
            mask <<= 1;
        }
        IntIntIntValue iv(result);
        conn.sendObject(&iv);
    }
}

/** THREADMV command.
    Parameters: thread Id, target forum Id */
void
doThreadMove(TalkConnection& conn, IntArgBlock& args)
{
    // Parse args
    args.checkArgumentCount(2);
    int32_t topicId = toIntResult(args.getNext());
    int32_t forumId = toIntResult(args.getNext());

    // Do it
    ReconnectingGuard g(database_rc);
    Topic t(database_connection, topicId);
    if (!t.exists()) {
        conn.sendError("404 Not found");
        return;
    }

    // Check forums
    Forum src(t.forum());
    Forum dst(database_connection, forumId);
    if (src.getId() == forumId) {
        conn.sendSuccess("OK (no change)");
        return;
    }

    // Check permissions: must have delete permission on source forum, and write permission on target
    if (!conn.isAdmin()) {
        conn.checkPrivilege(src.deletePermissions().get());
        conn.checkPrivilege(dst.writePermissions().get());
    }

    // Do it. Actual forum move is trivial, but we must update all sequence numbers and change
    // all message Ids for the NNTP side.
    IntList_t posts;
    t.messages().sort().by(MESSAGE_ROOT "*:header->seq").getResult(posts);
    for (uint32_t i = 0; i < posts.size(); ++i) {
        Message msg(database_connection, posts[i]);

        // NOTE: For a perfect mapping, we must actually make sure that the new sequence number does
        // not match ANY old sequence number, because otherwise the new message would have the same
        // generated message Id on the NNTP side. This could be trivially assured by choosing the
        // new one bigger than the old one, but that is not feasible (think moving a thread from a
        // 10000-post forum into a 10-post forum, leaving a 9990-post gap). Thus, we only avoid
        // clashes with the current and previous one.
        int32_t oldSeq1 = msg.sequenceNumber().get();
        int32_t oldSeq2 = msg.previousSequenceNumber().get();
        int32_t newSeq;
        do {
            newSeq = ++dst.lastMessageSequenceNumber();
        } while (newSeq == oldSeq1 || newSeq == oldSeq2);

        // Update the post's identifying information
        string_t rfcMsgId = msg.rfcMessageId().get();
        msg.removeRfcMessageId(database_connection, rfcMsgId);
        msg.previousRfcMessageId().set(rfcMsgId);
        msg.previousSequenceNumber().set(oldSeq1);
        msg.sequenceNumber().set(newSeq);
        msg.rfcMessageId().remove();
    }

    // Move the postings into the new forum
    src.messages().remove(t.messages()).storeTo(src.messages());
    dst.messages().merge(t.messages()).storeTo(dst.messages());

    // Move the thread
    if (t.isSticky()) {
        src.stickyTopics().moveTo(topicId, dst.stickyTopics());
    } else {
        src.topics().moveTo(topicId, dst.topics());
    }
    t.forumId().set(forumId);
    conn.sendSuccess("OK");
}

/** THREADRM command.
    Parameters: thread Id */
void
doThreadRemove(TalkConnection& conn, IntArgBlock& args)
{
    // Parse args
    args.checkArgumentCount(1);
    int32_t topicId = toIntResult(args.getNext());

    // Do it
    ReconnectingGuard g(database_rc);
    Topic t(database_connection, topicId);
    int32_t result;
    if (!t.exists()) {
        // Does not exist: assume success
        result = 0;
    } else {
        // Check delete permissions
        Forum f(t.forum());
        if (!conn.isAdmin()) {
            conn.checkPrivilege(f.deletePermissions().get());
        }

        // Do it
        t.remove();
        result = 1;
    }

    IntIntIntValue iv(result);
    conn.sendObject(&iv);
}
