/**
  *  \file planetscentral/talk/syntaxdb.cc
  */

#include <vector>
#include <cstdio>
#include "planetscentral/talk/syntaxdb.h"
#include "util/filepars.h"

namespace {
    class SyntaxParser : public FileParser {
     public:
        SyntaxParser(SyntaxDatabase& db);
        virtual void process(int line_id, string_t line);
        virtual void unparsed(int line_id, string_t line);
     private:
        void error(int line_id, const char* msg);

        SyntaxDatabase& db;
        std::vector<string_t> prefixes;
    };
}

SyntaxParser::SyntaxParser(SyntaxDatabase& db)
    : FileParser(";#"),
      db(db),
      prefixes()
{
    prefixes.push_back(string_t());
}

void
SyntaxParser::process(int line_id, string_t line)
{
    // Parse
    string_t::size_type n = line.find_first_of("={}");
    if (n == string_t::npos) {
        error(line_id, "syntax error");
        return;
    }

    // Key and value
    string_t key   = strTrim(string_t(line, 0, n));
    string_t value = strTrim(string_t(line, n+1));
    switch (line[n]) {
     case '=':
        if (key.empty()) {
            error(line_id, "missing key in assignment");
        } else if (!value.empty() && value[0] == '$') {
            // reference
            value.erase(0, 1);
            if (!db.has(value)) {
                error(line_id, "reference to non-existant key");
            } else {
                db.add(prefixes.back() + key, db.get(value));
            }
        } else {
            db.add(prefixes.back() + key, value);
        }
        break;

     case '{':
        if (key.empty() || !value.empty()) {
            error(line_id, "group must have form \"<name> {\"");
        } else {
            prefixes.push_back(prefixes.back() + key + ".");
        }
        break;

     case '}':
        if (!key.empty() || !value.empty()) {
            error(line_id, "group end must have form \"}\"");
        } else if (prefixes.size() <= 1) {
            error(line_id, "too many group terminators");
        } else {
            prefixes.pop_back();
        }
        break;
    }
}

void
SyntaxParser::unparsed(int /*line_id*/, string_t /*line*/)
{ }

void
SyntaxParser::error(int line_id, const char* msg)
{
    // We're not multithreaded here, so we can write without taking output_mutex,
    // saving a dependency
    std::printf("[syntaxdb] %s:%d: %s\n", getFileName().c_str(), line_id, msg);
}

/***************************** SyntaxDatabase ****************************/

SyntaxDatabase::SyntaxDatabase()
{ }

SyntaxDatabase::~SyntaxDatabase()
{ }

void
SyntaxDatabase::load(Stream& in)
{
    SyntaxParser(*this).parseFile(in);
}

void
SyntaxDatabase::add(const string_t& key, const string_t& value)
{
    data[strLCase(key)] = value;
}

bool
SyntaxDatabase::has(const string_t& key) const
{
    return data.find(strLCase(key)) != data.end();
}

string_t
SyntaxDatabase::get(const string_t& key) const
{
    std::map<string_t, string_t>::const_iterator it = data.find(strLCase(key));
    if (it != data.end()) {
        return it->second;
    } else {
        return string_t();
    }
}
